#!/usr/bin/env python -O
# ngrammodel.py: n-gram smoothing and perplexity estimation
# Kyle Gorman <gormanky@ohsu.edu> and Steven Bedrick <bedricks@ohsu.edu>
#
# This script defines classes representing different conditional
# probability estimators. Conditional frequency and probability
# distributions are consistently represented as dictionaries of
# dictionaries, with the the prefix as the outer key and the suffix
# as the inner key.

from __future__ import division

from collections import defaultdict

from ngrams import ngrams
from bitweight import BitWeight, INF


class MaximumLikelihoodNGramModel(object):

    """
    n-gram model with maximum likelihood estimation
    """

    def __init__(self, corpus, order):
        self.order = order
        self._compute_prob(corpus)

    # called in constructor
    def _compute_prob(self, corpus):
        """
        Compute a conditional frquency distribution and normalize it into
        BitWeight probabilities
        """
        # conditional frequency distribution representation
        self.freq = defaultdict(lambda: defaultdict(int))
        # populate it
        for (prefix, suffix) in ngrams(corpus, self.order):
            self.freq[prefix][suffix] += 1
        # initialize log2 probability distribution with default value (0)
        self.prob = defaultdict(lambda: defaultdict(lambda: BitWeight()))
        # populate seen probabilities
        for (prefix, suffixes) in self.freq.iteritems():
            # prefix is a tuple of tokens, sffixes is a dictionary
            # containing token: count key/value pairs
            denominator = BitWeight(sum(suffixes.values()))
            # optimize dictionary lookup; both of these are pointers, so
            # modifying pdist (as we will) modifies self.prob too
            fdist = self.freq[prefix]
            pdist = self.prob[prefix]
            # normalize, using fdist[suffix], the full n-gram's count
            for suffix in suffixes:
                # BitWeight class implements division as log-space
                # subtraction so as to forestall underflow
                pdist[suffix] = BitWeight(fdist[suffix]) / denominator

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    def perplexity(self, corpus):
        """
        Compute per-token perplexity. The cross-entropy of p with respect
        to q is defined as: 

        H(p, q)-\sum_x p(x) log_2 q(x)

        In the case of language modeling, q is the estimated probability
        distribution and p is the (MLE) probability of each n-gram in the 
        held-out data. Then, the perplexity of p w.r.t. q is simply

        PPX(p, q) = 2^{H / Z}

        where Z is the number of tokens in p.
        """
        # frequency of observations over corpus
        fx = defaultdict(int)
        # number of observations in corpus
        Z = 0
        for (prefix, suffix) in ngrams(corpus, self.order):
            Z += 1
            fx[(prefix, suffix)] += 1
        # collect entropy
        H = BitWeight(1.)
        for ((prefix, suffix), f_x) in fx.iteritems():
            H *= f_x * self.prob[prefix][suffix]
        # divide to get per-token entropy, exponentiate to get perplexity
        return 2 ** (H.bw / Z)


class WittenBellNGramModel(MaximumLikelihoodNGramModel):

    """
    n-gram model with Witten-Bell estimation
    """

    def _compute_prob(self, corpus):
        # collect lower-order MLE models
        models = [MaximumLikelihoodNGramModel(corpus, order) for
                  order in xrange(1, 1 + self.order)]
        # compute lambdas (models[0].prob: unigram model, (): null suffix)
        N = len(models[0].prob[()])           # num types - models[0] is a unigram model, and so the "blank" history of "()" is shared by all tokens
        T = sum(models[0].freq[()].values())  # num tokens 
        # handles n-grams with OOV prefixes and initializes internal
        # defaultdicts using the "zerogram" probability, modeled using a
        # uniform distribution over the unigram vocabulary
        defdef_factory = lambda N_sv=N, T_sv=T: BitWeight(N_sv / (N_sv + T_sv))
        self.prob = defaultdict(lambda: defaultdict(defdef_factory))
        
        # accumulate from there
        for model in models[1:]:
            for (prefix, suffixes) in model.prob.iteritems():
                # compute lambdas
                # types following this prefix; N(w_{i-1}) for Columbia
                N = len(suffixes)
                # tokens following this prefix; C(w_{i - 1}) for Columbia
                T = sum(model.freq[prefix].values())
                # smoothing lambda
                Lambda = BitWeight(T / (N + T))
                one_minus_Lambda = BitWeight(1. - Lambda.to_real)
                # optimize dictionary lookup
                mod_prefix_pdist = model.prob[prefix]
                cur_prefix_pdist = self.prob[prefix]
                # handles n-grams with seen prefixes but OOV suffixes;
                # same as inner probability except for a null MLE term
                bkf_term = one_minus_Lambda * cur_prefix_pdist[None]
                cur_prefix_pdist.default_factory = \
                    lambda bkf_term_sv=bkf_term: bkf_term_sv
                for (suffix, suffix_prob) in mod_prefix_pdist.iteritems():
                    # handles previously observed n-grams
                    mle_term = Lambda * suffix_prob
                    bkf_term = one_minus_Lambda * cur_prefix_pdist[suffix]
                    cur_prefix_pdist[suffix] += mle_term + bkf_term

if __name__ == '__main__':
    # Example Usage:
    # both models want a "corpus" for training
    # i.e., a list of lists of tokens:
    # [["This", "is", "the", "first", "sentence", "."], ["This", "is", "another", "one", "."]]
    # 
    # The models provide a .prob[] dictionary, which maps a hitory tuple to a defaultdict 
    # of word probabilities. Give the defaultdicts a word, and it will give you the model's
    # estimation of that word's likelihood given that history. 
    # 
    # The probability will be provided as a BitWeight, so as to avoid underflow.

    pride_corpus = [l.strip().upper().split() for l in open("pride_prejudice_ptb_tokenized.txt")]

    print "There are ",len(pride_corpus), "lines in the corpus."

    # make an MLE model:
    m = MaximumLikelihoodNGramModel(pride_corpus, 2) # make a bigram model
    obliged_probs = m.prob[('OBLIGED',)] # tells us about about probabilities of bigrams beginning with "OBLIGED"
    print(obliged_probs['TO']).to_real
    
    # no need to store the history dictionary into an intermediate variable; we can just access it directly:
    print(m.prob[('OBLIGED',)]['MONKEY']).to_real # should be zero- the bigram "OBLIGED MONKEY" does not appear.
    
    # make a W-B model
    m2 = WittenBellNGramModel(pride_corpus, 2) # make a W-B smoothed bigram model
    print(m2.prob[('OBLIGED',)]['TO']).to_real
    print(m2.prob[('OBLIGED',)]['MONKEY']).to_real # should be small, not zero

    # for (prefix, suffix) in ngrams(pride_corpus[1:5], 2):
    #     print prefix
    
    # ask about starting probabilities:
    print(m2.prob[('<S_0>',)]['WHEN']) # probability of a sentence starting with "WHEN"
    print(m2.prob[('<S_0>',)]['ZEBRAS']) # probability of a sentence starting with "ZEBRAS"
    
    # what are all of the observed sentence beginnings?
    print(m.prob[('<S_0>',)].keys())